/*
 * Template.h
 *
 *  Created on: Mar 5, 2016
 *      Author: kesalin
 */

#ifndef TEMPLATE_H_
#define TEMPLATE_H_

#include <stdio.h>
#include "Base.h"

// Helper types Small and Big - guarantee that sizeof(Small) < sizeof(Big)
//
template <class T, class U>
struct ConversionHelper
{
    typedef char Small;
    struct Big { char dummy[2]; };
    static Big   Test(...);
    static Small Test(U);
    static T & MakeT();
};

// class template Conversion
// Figures out the conversion relationships between two types
// Invocations (T and U are types):
// exists: returns (at compile time) true if there is an implicit conversion
// from T to U (example: Derived to Base)
// Caveat: might not work if T and U are in a private inheritance hierarchy.
//
template <class T, class U>
struct Conversion
{
    typedef ConversionHelper<T, U> H;
    enum {
        exists = sizeof(typename H::Small) == sizeof((H::Test(H::MakeT())))
    };
    enum { exists2Way = exists && Conversion<U, T>::exists };
    enum { sameType = false };
};

template <class T>
class Conversion<T, T>
{
public:
    enum { exists = true, exists2Way = true, sameType = true };
};

#ifndef SUPERSUBCLASS
#define SUPERSUBCLASS(Super, Sub) \
    (Conversion<Sub, Super>::exists && !Conversion<Super, void*>::sameType)
#endif

template<class T, bool isTypeOfBase>
struct ProcessFunc
{
    void operator()(T obj)
    {
        printf("It's not type of Base.\n");
    }
};

template<class T>
struct ProcessFunc<T, true>
{
    void operator()(T obj)
    {
        printf("It's type of Base. GetName: %s\n", obj->GetName());
    }
};

template <class T>
class Holder {
public:
    Holder(T obj)
        : mValue(obj)
    {}

    void Print()
    {
        ProcessFunc<T, SUPERSUBCLASS(Base*, T) > func;
        func(mValue);
    }

private:
    T mValue;
};

#endif /* TEMPLATE_H_ */
